from __future__ import annotations
from typing import Any

from typeguard import typechecked
from getpass import getpass
import os

from termcolor import colored

from mpqp.execution.connection.env_manager import (
    save_env_variable,
    get_env_variable,
    load_env_variables,
)
from mpqp.tools.errors import QLMRemoteExecutionError

QLM_connection = None


@typechecked
def config_qlm_account(username: str, password: str, global_config: bool) -> bool:
    """Configures and saves locally QLM account's information.

    Args:
        username: QLM username.
        password: QLM password.
        global_config: If True, this QLM account will be configured to work even
            outside MPQP.

    Raises:
        QLMRemoteExecutionError
    """
    # store the username and password in environment variables QLM_USER and QLM_PASSWD in .mpqp
    prev_user = get_env_variable("QLM_USER")
    prev_pass = get_env_variable("QLM_PASSWD")
    prev_configure = get_env_variable("QLM_CONFIGURED")
    if prev_configure == "":
        prev_configure = "False"

    save_env_variable("QLM_USER", username)
    if not global_config:
        save_env_variable("QLM_PASSWD", password)

    try:
        if global_config:
            print("we are in the global part")
            netrc_path = os.path.expanduser("~") + "/.netrc"
            # if file doesn't exist, create it, or overwrite the credentials in the ~/.netrc file
            with open(netrc_path, "w") as file:
                file.write(
                    "machine qlm35e.neasqc.eu\nlogin "
                    + username
                    + "\npassword "
                    + password
                )
            # Set the permissions to '0600' (equivalent to 'chmod og-rw')
            os.chmod(netrc_path, 0o600)

        from qat.qlmaas import QLMaaSConnection  # type: ignore

        c = QLMaaSConnection(
            hostname="qlm35e.neasqc.eu",
            port=443,
            authentication="password",
            timeout=2500,
        )
        c.create_config()

    except Exception as err:
        save_env_variable("QLM_USER", prev_user)
        save_env_variable("QLM_PASSWD", prev_pass)
        save_env_variable("QLM_CONFIGURED", prev_configure)
        if "Invalid credential" in str(err):
            return False
        raise QLMRemoteExecutionError(
            "Error when saving the account.\nTrace: " + str(err)
        )
    return True


def setup_qlm_account() -> tuple[str, list[Any]]:
    """Setups the QLM account, by looking at the existing configuration, asking
    for username/password and updating the current account."""
    already_configured = get_env_variable("QLM_CONFIGURED") == "True"

    if already_configured:
        decision = input(
            "An QLM account is already configured. Do you want to update it? [y/N] "
        )
        if decision.lower().strip() != "y":
            return "Canceled.", []

    username = input("Enter your QLM username: ")
    password = getpass("Enter your QLM password (hidden): ")
    global_decision = input(
        "Do you want to configure this account only for MPQP? [y/N] "
    )
    global_config = True if global_decision.lower().strip() != "y" else False

    connection_success = config_qlm_account(username, password, global_config)
    if connection_success:
        save_env_variable("QLM_CONFIGURED", "True")

    if connection_success:
        return "QLM account correctly configured", []
    else:
        print(colored("Wrong credential", "red"))
        getpass("Press 'Enter' to continue")
        return "", []


def get_all_job_ids() -> list[str]:
    """Retrieves from the remote QLM all the job-ids associated with this account.

    Example:
        >>> get_all_job_ids()
        ['Job144361', 'Job144360', 'Job144359', 'Job144358', 'Job144357', 'Job143334', 'Job143333', 'Job143332',
        'Job141862', 'Job141861', 'Job141722', 'Job141720', 'Job141716', 'Job141715', 'Job141712', 'Job19341']

    Returns:
        List of all job-ids associated with this account.
    """

    connection = get_QLMaaSConnection()
    return [job_info.id for job_info in connection.get_jobs_info()]


def get_QLMaaSConnection():
    """Connects to the QLM and returns the QLMaaSConnection. If the connection
    was already established, we only return the one stored in global variable,
    otherwise we instantiate a new QLMaaSConnection."""

    global QLM_connection
    if QLM_connection is None:
        loaded = load_env_variables()
        if not loaded:
            raise IOError("Could not load environment variables from ~/.mpqp")

        if get_env_variable("QLM_CONFIGURED") == "False":
            raise QLMRemoteExecutionError(
                "Error when instantiating QLMaaSConnection. No QLM account configured."
            )

        from qat.qlmaas.connection import QLMaaSConnection

        QLM_connection = QLMaaSConnection(
            hostname="qlm35e.neasqc.eu",
            port=443,
            authentication="password",
            timeout=2500,
        )
    return QLM_connection



from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

with open("LICENSE", "r") as f:
    license = f.readline()

with open("requirements.txt", "r") as f:
    requirements = f.readlines()

 

setup(
    name="mpqp",
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    description="Facilitate quantum algorithm development and execution, regardless of the hardware, with MPQP",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license=license,
    author="ColibriTD",
    author_email="quantum@colibritd.com",
    install_requires=["wheel"] + requirements,
    packages=find_packages(include=["mpqp*"]),
    entry_points={
        "console_scripts": [
            "setup_connections = mpqp.execution.connection.setup_connections:main_setup",
        ]
    },
    package_data={"mpqp.qasm.header_codes": ["*.qasm", "*.inc"]},
)



""":class:`Measure<mpqp.core.instruction.measurement.measure.Measure>` is the
base class for measurements. It regroups all the attributes and methods
common to all types of measurements we support.

A measurement can be of two types:
:class:`BasisMeasure<mpqp.core.instruction.measurement.basis_measure.BasisMeasure>`
or
:class:`ExpectationMeasure<mpqp.core.instruction.measurement.expectation_value.ExpectationMeasure>`,
described in details bellow."""

from __future__ import annotations

from abc import ABC
from typing import Optional

from typeguard import typechecked, TypeCheckError

from mpqp.core.instruction import Instruction


@typechecked
class Measure(Instruction, ABC):
    """Class representing a generic measure of quantum state generated by a
    quantum circuit.

    This class is used to regroup attributes and methods shared by all different
    types of measures.

    We distinguish two types of measures:

    - Basis measurement (measure some qubits in a specific basis, sample
      mode, or retrieve the StateVector when shots is equal to zero)
    - Expectation value (use of an observable, exact or sample mode)

    Args:
        targets: List of indices referring to the qubits on which the measure
            will be applied.
        shots: Number of times the circuit should be run, each of these times is
            called a shot.
        label: Label used to identify the measure.
    """

    def __init__(
        self,
        targets: list[int],
        shots: int = 0,
        label: Optional[str] = None,
    ):
        if shots < 0:
            raise TypeCheckError(
                f"Negative number of shot makes no sense, given {shots}"
            )
        super().__init__(targets, label)
        self.shots = shots
        """See parameter description."""
        self.label = label
        """See parameter description."""



from typing import Optional

from typeguard import typechecked

from mpqp.core.instruction.gates.gate import Gate
from mpqp.core.instruction.gates.gate_definition import (
    UnitaryMatrix,
    KrausRepresentation,
    PauliDecomposition,
)
from mpqp.core.languages import Language
from qiskit.circuit import Parameter


@typechecked
class CustomGate(Gate):
    """Custom gates allow you to define your own gates.

    Args:
        definition: The matrix (this is the only way supported to now) semantics of the gate.
        targets: The qubits on which the gate operates.
        label: The label of the gate. Defaults to None.
    """

    def __init__(
        self, definition: UnitaryMatrix, targets: list[int], label: Optional[str] = None
    ):
        self.matrix = definition.matrix
        """See parameter description."""
        super().__init__(targets, label)

    def to_matrix(self):
        return self.matrix

    def to_other_language(
        self,
        language: Language = Language.QISKIT,
        qiskit_parameters: Optional[set[Parameter]] = None,
    ):
        if qiskit_parameters is None:
            qiskit_parameters = set()
        return super().to_other_language(language, qiskit_parameters)

    def decompose(self):
        """Returns the circuit made of native gates equivalent to this gate.

        6M-TODO refine this doc and implement
        """
        from mpqp.core.circuit import QCircuit

        return QCircuit(self.nb_qubits)


@typechecked
class KrausGate(CustomGate):
    """6M-TODO"""
    def __init__(
        self,
        definition: KrausRepresentation,
        targets: list[int],
        label: Optional[str] = None,
    ):
        self.kraus_representation = definition
        """See parameter description."""
        CustomGate.__init__(self, UnitaryMatrix(definition.to_matrix()), targets, label)


@typechecked
class PauliDecompositionGate(CustomGate):
    """6M-TODO"""
    def __init__(
        self,
        definition: PauliDecomposition,
        targets: list[int],
        label: Optional[str] = None,
    ):
        self.pauli_decomposition = definition
        """See parameter description."""
        CustomGate.__init__(self, UnitaryMatrix(definition.to_matrix()), targets, label)



from __future__ import annotations

from abc import abstractmethod, ABC
from warnings import warn

import numpy as np
from sympy import Expr
from typeguard import typechecked
from numbers import Complex

from mpqp.tools.maths import is_unitary, matrix_eq
from mpqp.tools.generics import Matrix, one_lined_repr


@typechecked
class GateDefinition(ABC):
    """A class used to handle the definition of a Gate.

    A quantum gate can be defined in several ways, and this class allows us to
    define it as we prefer. It also handles the translation from one definition
    to another.

    This said, for now only one way of defining the gates is supported, using
    their matricial semantics.

    Example:
        >>> gate_matrix = np.array([[0, 0, 0, 1], [0, 1, 0, 0], [1, 0, 0, 0], [0, 0, 1, 0]])
        >>> gate_definition = UnitaryMatrix(gate_matrix)
        >>> custom_gate = CustomGate(gate_definition)
    """

    """ TODO: put this back once we implement the other definitions
    This class permit to define a gate in 4 potential ways:
        1. the unitary matrix defining the gate
        2. a combination of several other gates
        3. a combination of Kraus operators
        4. the decomposition of the gate in the Pauli basis (only possible for LU gates)
    """

    @abstractmethod
    def to_matrix(self) -> Matrix:
        """Returns the matrix corresponding to this gate definition."""

    @abstractmethod
    def subs(
        self,
        values: dict[Expr | str, Complex],
        remove_symbolic: bool = False,
        disable_symbol_warn: bool = False,
    ) -> GateDefinition:
        pass

    # @abstractmethod
    def to_kraus_representation(self) -> KrausRepresentation:
        """6M-TODO"""
        raise NotImplementedError()

    # @abstractmethod
    def to_pauli_decomposition(self) -> PauliDecomposition:
        """6M-TODO"""
        raise NotImplementedError()

    def is_equivalent(self, other: GateDefinition) -> bool:
        """Determines if this definition is equivalent to the other.

        Args:
            other: The definition we want to know if it is equivalent.

        Example:
            >>> d1 = UnitaryMatrix(np.array([[1, 0], [0, -1]]))
            >>> d2 = UnitaryMatrix(np.array([[2, 0], [0, -2.0]]) / 2)
            >>> d1.is_equivalent(d2)
            True
        """
        return matrix_eq(self.to_matrix(), other.to_matrix())

    def inverse(self) -> GateDefinition:
        """Compute the inverse of the gate.

        Returns:
            A GateDefinition representing the inverse of the gate defined.

        Example:
            >>> UnitaryMatrix(np.array([[1, 0], [0, -1]])).inverse()
            array([[ 1.,  0.],
                   [-0., -1.]])
        """
        mat = self.to_matrix()

        if not all(
            isinstance(elt.item(), Complex) for elt in np.nditer(mat, ["refs_ok"])  # type: ignore
        ):
            raise ValueError("Cannot invert arbitrary gates using symbolic variables")
        return UnitaryMatrix(np.linalg.inv(mat))  # type:ignore


class KrausRepresentation(GateDefinition):
    """# 6M-TODO : implement and comment"""

    def __init__(self):
        self.pp = 1


class PauliDecomposition(GateDefinition):
    """# 6M-TODO : implement and comment"""

    def __init__(self):
        self.pp = 1


@typechecked
class UnitaryMatrix(GateDefinition):
    """Definition of a gate using it's matrix.

    Args:
        definition: Matrix defining the unitary gate.
        disable_symbol_warn: Boolean used to enable/disable warning concerning
            unitary checking with symbolic variables.
    """

    def __init__(self, definition: Matrix, disable_symbol_warn: bool = False):
        if any(isinstance(elt, Expr) for _, elt in np.ndenumerate(definition)):
            if not disable_symbol_warn:
                # 3M-TODO: can we improve this situation ?
                warn("Cannot ensure that a operator defined with variables is unitary.")
        elif not is_unitary(definition):
            raise ValueError(
                "Matrices defining gates have to be unitary. It is not the case"
                f" for\n{definition}"
            )
        self.matrix = definition
        """See parameter :attr:`definition`'s description."""

    def to_matrix(self) -> Matrix:
        return self.matrix

    def to_kraus_representation(self) -> KrausRepresentation:
        """6M-TODO to implement"""
        ...

    def to_pauli_decomposition(self) -> PauliDecomposition:
        """6M-TODO to implement"""
        ...

    def subs(
        self,
        values: dict[Expr | str, Complex],
        remove_symbolic: bool = False,
        disable_symbol_warn: bool = False,
    ):
        """Substitute some symbolic variables in the definition by complex values.

        Args:
            values: Mapping between the symbolic variables and their complex
                attributions.
            remove_symbolic: Some values such as pi are kept symbolic during
                circuit manipulation for better precision, but must be replaced
                by their complex counterpart for circuit execution, this
                arguments fills that role. Defaults to False.
            disable_symbol_warn: This method returns a :class:`UnitaryMatrix`,
                which raises a warning in case the matrix used to build it has
                symbolic variables. This is because this class performs
                verifications on the matrix to ensure it is indeed unitary, but
                those verifications cannot be done on symbolic variables. This
                argument disables this check because in some contexts, it is
                undesired. Defaults to False.
        """

        def mapping(val: Expr | Complex) -> Expr | Complex:
            def caster(v: Expr | Complex) -> Expr | Complex:
                # problem between pyright and abstract numeric types ?
                # "complex" cannot be assigned to return type "Complex"
                return (
                    complex(v) if remove_symbolic else v
                )  # pyright: ignore[reportReturnType]

            # the types in sympy are relatively badly handled
            # Argument of type "Unknown | Basic | Expr" cannot be assigned to parameter "v" of type "Expr | Complex"
            return (
                caster(val.subs(values))  # pyright: ignore[reportArgumentType]
                if isinstance(val, Expr)
                else val
            )

        matrix = self.to_matrix()
        otype = (
            complex
            if remove_symbolic
            or not any(isinstance(val, Expr) for _, val in np.ndenumerate(matrix))
            else object
        )

        return UnitaryMatrix(
            np.vectorize(mapping, otypes=[otype])(matrix), disable_symbol_warn
        )

    def __repr__(self) -> str:
        return f"UnitaryMatrix({one_lined_repr(getattr(self, 'matrix', ''))})"



######################################
# Copyright(C) 2021 - 2023 ColibrITD
#
# Developers :
#  - Hamza JAFFALI < hamza.jaffali@colibritd.com >
#  - Karla BAUMANN < karla.baumann@colibritd.com >
#  - Henri de BOUTRAY < henri.de.boutray@colibritd.com >
#
# Version : 0.1
#
# This file is part of QUICK.
#
# QUICK can not be copied and / or distributed without the express
# permission of ColibrITD
#
######################################
from __future__ import annotations

import math
from numbers import Complex
from textwrap import dedent
from typing import Optional

import numpy as np
import numpy.typing as npt
from typeguard import typechecked

from mpqp.execution.devices import AvailableDevice
from mpqp.tools.errors import ResultAttributeError

from .job import Job, JobType


@typechecked
class StateVector:
    """
    Class representing the state vector of a multi-qubit quantum system.

    Args:
        vector: List of amplitudes defining the state vector.
        nb_qubits: Number of qubits of the state.
        probabilities: List of probabilities associated with the state vector.

    Example:
        >>> state_vector = StateVector(np.array([1, 1, 1, -1])/2, 2)
        >>> state_vector.probabilities
        array([0.25, 0.25, 0.25, 0.25])
        >>> print(state_vector)
        State vector: [ 0.5  0.5  0.5 -0.5]
        Probabilities: [0.25 0.25 0.25 0.25]
        Number of qubits: 2

    """

    def __init__(
        self,
        vector: list[Complex] | npt.NDArray[np.complex64],
        nb_qubits: Optional[int] = None,
        probabilities: Optional[list[float] | npt.NDArray[np.float32]] = None,
    ):
        if len(np.asarray(vector)) == 0:
            raise ValueError("vector should not be empty")

        self.vector: npt.NDArray[np.complex64] = np.array(vector, dtype=complex)

        self.nb_qubits = (
            int(math.log(len(vector), 2)) if nb_qubits is None else nb_qubits
        )
        """See parameter description."""
        self.probabilities = (
            abs(self.vector) ** 2 if probabilities is None else np.array(probabilities)
        )
        """See parameter description."""

    @property
    def amplitudes(self):
        """Return the amplitudes of the state vector"""
        return self.vector

    def __str__(self):
        return f"""
        State vector: {self.vector}
        Probabilities: {self.probabilities}
        Number of qubits: {self.nb_qubits}"""


@typechecked
class Sample:
    """
    Class representing a sample, which contains the result of the experiment concerning a specific basis state
    of the Hilbert space.

    Args:
        nb_qubits: Number of qubits of the quantum system of the experiment.
        probability: Probability of measuring the basis state associated to this
            sample.
        index: Index in decimal notation representing the basis state.
        count: Number of times this basis state was measured during the
            experiment.
        bin_str: String representing the basis state in binary notation.

    Examples:
        >>> print(Sample(3, index=3, count=250, bin_str="011"))
        State: 011, Index: 3, Count: 250, Probability: None

        >>> print(Sample(4, index=6, probability=0.5))
        State: 0110, Index: 6, Count: None, Probability: 0.5

        >>> print(Sample(5, bin_str="01011", count=1234))
        State: 01011, Index: 11, Count: 1234, Probability: None
    """

    def __init__(
        self,
        nb_qubits: int,
        probability: Optional[float] = None,
        index: Optional[int] = None,
        count: Optional[int] = None,
        bin_str: Optional[str] = None,
    ):
        self.nb_qubits = nb_qubits
        """See parameter description."""
        self.count = count
        """See parameter description."""
        self.probability = probability
        """See parameter description."""
        self.bin_str: str
        """See parameter description."""
        self.index: int
        """See parameter description."""

        if index is None:
            if bin_str is None:
                raise ValueError(
                    "At least one of `bin_str` and `index` arguments is necessary"
                )
            else:
                self.index = int(bin_str, 2)
                self.bin_str = bin_str
        else:
            computed_bin_str = bin(index)[2:].zfill(self.nb_qubits)
            if bin_str is None:
                self.index = index
                self.bin_str = computed_bin_str
            else:
                if computed_bin_str == bin_str:
                    self.index = index
                    self.bin_str = bin_str
                else:
                    raise ValueError(
                        f"The value of bin_str {bin_str} doesn't match with the"
                        f" index provided {index}"
                    )

    def __str__(self):
        str1 = "State: " + str(self.bin_str) + ", Index: " + str(self.index)
        str2 = ", Count: " + str(self.count) + ", Probability: " + str(self.probability)
        return str1 + str2

    def __repr__(self):
        str1 = "State: " + str(self.bin_str) + ", Index: " + str(self.index)
        str2 = ", Count: " + str(self.count) + ", Probability: " + str(self.probability)
        return str1 + str2


@typechecked
class Result:
    """Result associated to a submitted job.

    The data type in a result depends on the job type, according to the
    following chart:

    +-------------+--------------+
    | Job Type    | Data Type    |
    +=============+==============+
    | OBSERVABLE  | float        |
    +-------------+--------------+
    | SAMPLE      | list[Sample] |
    +-------------+--------------+
    | STATE_VECTOR| StateVector  |
    +-------------+--------------+

    Args:
        job: Type of the job related to this result.
        data: Data of the result, can be an expectation value (float), a
            StateVector, or a list of sample depending on the job_type.
        error: Information about the error or the variance in the measurement.
        shots: Number of shots of the experiment (equal to zero if the exact
            value was required).

    Examples:
        >>> print(Result(Job(), StateVector(np.array([1, 1, 1, -1])/2, 2), 0, 0))
        State vector: [ 0.5  0.5  0.5 -0.5]
        Probabilities: [0.25 0.25 0.25 0.25]
        Number of qubits: 2

        >>> print(Result(Job(), [
        ...     Sample(2, index=0, count=250)
        ...     Sample(2, index=3, count=250)
        ... ], 0.034, 500))
        Counts: [250, 250]
        Probabilities: [0.5 0.5]
        State: 00, Index: 0, Count: 250, Probability: None
        State: 11, Index: 3, Count: 250, Probability: None
        Error: 0.034

        >>> print(Result(Job(), -3.09834, 0.021, 2048))
        Expectation value: -3.09834
        Error: 0.021
    """

    # 3M-TODO: in this class, there is a lot of manual type checking, this is an
    #  anti-pattern in my opinion, it should probably be fixed using subclasses

    def __init__(
        self,
        job: Job,
        data: float | StateVector | list[Sample],
        error: Optional[float] = None,
        shots: int = 0,
    ):
        self.job = job
        """See parameter description."""
        self._expectation_value = None
        self._state_vector = None
        self._probabilities = None
        self._counts = None
        self._samples = None
        self.shots = shots
        """See parameter description."""
        self.error = error
        """See parameter description."""

        # depending on the type of job, fills the result info from the data in parameter
        if job.job_type == JobType.OBSERVABLE:
            if not isinstance(data, float):
                raise TypeError(
                    "Wrong type of data in the result. "
                    "Expecting float for expectation value of an observable"
                )
            else:
                self._expectation_value = data
        elif job.job_type == JobType.STATE_VECTOR:
            if not isinstance(data, StateVector):
                raise TypeError(
                    "Wrong type of data in the result. Expecting StateVector"
                )
            else:
                self._state_vector = data
                self._probabilities = data.probabilities
        elif job.job_type == JobType.SAMPLE:
            if not isinstance(data, list):
                raise TypeError(
                    "Wrong type of data in the result (not a list). Expecting list of Sample"
                )
            if self.job.measure is None:
                raise ValueError(
                    f"{self.job=} has no measure, making the counting impossible"
                )
            self._samples = data
            is_counts = all([sample.count is not None for sample in data])
            is_probas = all([sample.probability is not None for sample in data])
            if is_probas:
                probas = [0.0] * (2**self.job.measure.nb_qubits)
                for sample in data:
                    probas[sample.index] = sample.probability
                self._probabilities = np.array(probas, dtype=float)

                counts = [
                    int(count)
                    for count in np.round(self.job.measure.shots * self._probabilities)
                ]
                self._counts = counts
                for sample in self._samples:
                    sample.count = self._counts[sample.index]
            elif is_counts:
                counts: list[int] = [0] * (2**self.job.measure.nb_qubits)
                for sample in data:
                    assert sample.count is not None
                    counts[sample.index] = sample.count
                self._counts = counts
                # if is_counts shots != 0
                assert shots != 0
                self._probabilities = np.array(counts, dtype=float) / self.shots
                for sample in self._samples:
                    sample.probability = self._probabilities[sample.index]
            else:
                raise ValueError(
                    f"For {JobType.SAMPLE.name} jobs, all samples must contain"
                    " either `count` or `probability` (and the non-None "
                    "attribute amongst the two must be the same in all samples)."
                )
        else:
            raise ValueError(f"{job.job_type} not handled")

    @property
    def device(self) -> AvailableDevice:
        """Device on which the job of this result was run"""
        return self.job.device

    @property
    def expectation_value(self) -> float:
        """Get the expectation value stored in this result"""
        if self.job.job_type != JobType.OBSERVABLE:
            raise ResultAttributeError(
                f"Job type: {self.job.job_type.name} but cannot get expectation"
                " value if the job type is not OBSERVABLE."
            )
        assert self._expectation_value is not None
        return self._expectation_value

    @property
    def amplitudes(self) -> npt.NDArray[np.complex64]:
        """Get the amplitudes of the state of this result"""
        if self.job.job_type != JobType.STATE_VECTOR:
            raise ResultAttributeError(
                "Cannot get amplitudes if the job was not of type STATE_VECTOR"
            )
        assert self._state_vector is not None
        return self._state_vector.amplitudes

    @property
    def state_vector(self) -> StateVector:
        """Get the state vector of the state associated with this result"""
        if self.job.job_type != JobType.STATE_VECTOR:
            raise ResultAttributeError(
                "Cannot get state vector if the job was not of type STATE_VECTOR"
            )
        assert self._state_vector is not None
        return self._state_vector

    @property
    def samples(self) -> list[Sample]:
        """Get the list of samples of the result"""
        if self.job.job_type != JobType.SAMPLE:
            raise ResultAttributeError(
                "Cannot get samples if the job was not of type SAMPLE"
            )
        assert self._samples is not None
        return self._samples

    @property
    def probabilities(self) -> npt.NDArray[np.float32]:
        """Get the list of probabilities associated with this result"""
        if self.job.job_type not in (JobType.SAMPLE, JobType.STATE_VECTOR):
            raise ResultAttributeError(
                "Cannot get probabilities if the job was not of"
                " type SAMPLE or STATE_VECTOR"
            )
        assert self._probabilities is not None
        return self._probabilities

    @property
    def counts(self) -> list[int]:
        """Get the list of counts for each sample of the experiment"""
        if self.job.job_type != JobType.SAMPLE:
            raise ResultAttributeError(
                "Cannot get counts if the job was not of type SAMPLE"
            )

        assert self._counts is not None
        return self._counts

    def __str__(self):
        header = f"Result: {type(self.device).__name__}, {self.device.name}"
        if self.job.job_type == JobType.SAMPLE:
            samples_str = ("\n" + " " * 16).join(map(str, self.samples))
            cleaned_probas = str(self._probabilities).replace("\n", " ")
            return header + dedent(
                f"""
                Counts: {self._counts}
                Probabilities: {cleaned_probas}
                {samples_str}
                Error: {self.error}\n\n"""
            )
        if self.job.job_type == JobType.STATE_VECTOR:
            return f"""{header}\n{self._state_vector}\n\n"""
        if self.job.job_type == JobType.OBSERVABLE:
            return header + dedent(
                f"""
                Expectation value: {self.expectation_value}
                Error/Variance: {self.error}\n\n"""
            )
        raise NotImplementedError(
            f"Job type {self.job.job_type} not implemented for __str__ method"
        )


@typechecked
class BatchResult:
    """Class used to handle several Result instances.

    Args:
        results: List of results.

    Example:
        >>> result1 = Result(
        ...     Job(JobType.STATE_VECTOR,QCircuit(0),ATOSDevice.MYQLM_PYLINALG),
        ...     StateVector(np.array([1, 1, 1, -1])/2, 2),
        ...     0,
        ...     0
        ... )
        >>> result2 = Result(
        ...     Job(
        ...         JobType.SAMPLE,
        ...         QCircuit(0),
        ...         ATOSDevice.MYQLM_PYLINALG,
        ...         BasisMeasure([0,1],shots=500)
        ...     ),
        ...     [Sample(2, index=0, count=250), Sample(2, index=3, count=250)],
        ...     0.034,
        ...     500)
        >>> result3 = Result(
        ...     Job(JobType.OBSERVABLE,QCircuit(0),ATOSDevice.MYQLM_PYLINALG),
        ...     -3.09834,
        ...     0.021,
        ...     2048
        ... )
        >>> batch_result = BatchResult([result1, result2, result3])
        >>> print(batch_result)
        BatchResult: 3 results
        Result: ATOSDevice, MYQLM_PYLINALG
        State vector: [ 0.5+0.j  0.5+0.j  0.5+0.j -0.5+0.j]
                Probabilities: [0.25 0.25 0.25 0.25]
                Number of qubits: 2
        Result: ATOSDevice, MYQLM_PYLINALG
        Counts: [250, 0, 0, 250]
        Probabilities: [0.5 0.  0.  0.5]
        State: 00, Index: 0, Count: 250, Probability: 0.5
        State: 11, Index: 3, Count: 250, Probability: 0.5
        Error: 0.034
        Result: ATOSDevice, MYQLM_PYLINALG
        Expectation value: -3.09834
        Error/Variance: 0.021
        >>> print(batch_result[0])
        Result: ATOSDevice, MYQLM_PYLINALG
            State vector: [ 0.5+0.j  0.5+0.j  0.5+0.j -0.5+0.j]
            Probabilities: [0.25 0.25 0.25 0.25]
            Number of qubits: 2
    """

    def __init__(self, results: list[Result]):
        self.results = results
        """See parameter description."""

    def __str__(self):
        header = f"BatchResult: {len(self.results)} results\n"
        body = "".join(map(str, self.results))
        return header + body

    def __repr__(self):
        return str(self)

    def __getitem__(self, index: int):
        return self.results[index]



# pyright: reportUnusedImport=false
import numpy as np
from . import QCircuit, Barrier, Language, Instruction
from .gates import (
    X,
    Y,
    Z,
    H,
    P,
    S,
    T,
    SWAP,
    U,
    Rx,
    Ry,
    Rz,
    Rk,
    CNOT,
    CZ,
    CRk,
    TOF,
    ControlledGate,
    Gate,
    ParametrizedGate,
    symbols,
    GateDefinition,
    KrausRepresentation,
    PauliDecomposition,
    CustomGate,
    UnitaryMatrix,
)
from .measures import (
    ComputationalBasis,
    Basis,
    HadamardBasis,
    VariableSizeBasis,
    BasisMeasure,
    ExpectationMeasure,
    Observable,
    Measure,
)
from .execution import (
    Result,
    StateVector,
    Sample,
    run,
    adjust_measure,
    submit,
    Job,
    JobStatus,
    JobType,
)
from .execution.devices import (
    ATOSDevice,
    IBMDevice,
    AWSDevice,
)
from .execution.vqa import minimize, Optimizer
from .execution.connection.qlm_connection import get_all_job_ids
from mpqp.execution.providers_execution.atos_execution import get_result_from_qlm_job_id
from .qasm import open_qasm_file_conversion_2_to_3, open_qasm_hard_includes

theta, k = symbols("θ k")  # type: ignore
obs = Observable(np.array([[0, 1], [1, 0]]))
circ = QCircuit([P(theta, 0), ExpectationMeasure([0], observable=obs, shots=1000)])



"""After the jobs are submitted, one can use the functions of this module to
retrieve the results from a job_id or the job directly, and list all job
attached to the configured accounts."""

from __future__ import annotations

from typeguard import typechecked

from mpqp.execution import Result
from mpqp.execution.connection.aws_connection import get_all_task_ids
from mpqp.execution.connection.qlm_connection import get_all_job_ids as qlm_ids
from mpqp.execution.connection.ibm_connection import get_all_job_ids as ibm_ids
from mpqp.execution.providers_execution.atos_execution import get_result_from_qlm_job_id
from mpqp.execution.providers_execution.aws_execution import (
    get_result_from_aws_task_arn,
)
from mpqp.execution.providers_execution.ibm_execution import get_result_from_ibm_job_id

from mpqp.execution.devices import IBMDevice, AWSDevice, ATOSDevice, AvailableDevice
from mpqp.execution.job import Job


@typechecked
def remote_result_from_id(job_id: str, device: AvailableDevice) -> Result:
    """
    Retrieve and parse a remote the result from a job_id and device. If the job is still running, it will wait until it
    is done.

    Examples:
        >>> print(remote_result_from_id('Job141933', ATOSDevice.QLM_LINALG))
        Result: ATOSDevice, QLM_LINALG
        Counts: [1017, 0, 0, 0, 983, 0, 0, 0]
        Probabilities: [0.5085 0.     0.     0.     0.4915 0.     0.     0.    ]
        State: 000, Index: 0, Count: 1017, Probability: 0.5085
        State: 100, Index: 4, Count: 983, Probability: 0.4915
        Error: 0.011181519941139355
        >>> print(remote_result_from_id(
        ...     'cm80pb1054sir2ck9i3g',
        ...     IBMDevice.IBMQ_QASM_SIMULATOR,
        ... ))
        Result: IBMDevice, IBMQ_QASM_SIMULATOR
        Expectation value: 1.6410799999999999
        Error/Variance: 1.24570724535
        >>> aws_task_id = (
        ...     'arn:aws:braket:us-east-1:752542621531:quantum-task/'
        ...     '6a46ae9a-d02f-4a23-b46f-eae43471bc22'
        ... )
        >>> print(remote_result_from_id(
        ...     aws_task_id,
        ...     AWSDevice.BRAKET_SV1_SIMULATOR,
        ... ))
        Result: AWSDevice, BRAKET_SV1_SIMULATOR
        Expectation value: 1.6635202030411578
        Error/Variance: None

    Args:
        job_id: Id used to identify the job on the remote device.
        device: Remote device on which the job was launched.

    Returns:
        A Result of the remote job identified by the job_id in parameter.
    """

    # if the device is not a remote one
    if not device.is_remote():
        raise ValueError(
            "Trying to retrieve a remote result while the device of the job was local."
        )

    # check the type of the device, and call the right job getter
    if isinstance(device, IBMDevice):
        return get_result_from_ibm_job_id(job_id)
    elif isinstance(device, ATOSDevice):
        return get_result_from_qlm_job_id(job_id)
    elif isinstance(device, AWSDevice):
        return get_result_from_aws_task_arn(job_id)
    else:
        raise NotImplementedError(
            f"The device {device.name} is not supported for remote features."
        )


@typechecked
def remote_result_from_job(job: Job) -> Result:
    """Retrieve and parse a remote the result from an mpqp job. If the job is
    still running, it will wait until it is done.

    Args:
        job: MPQP job for which we want to retrieve the result from the remote
            device.

    Returns:
        A Result of the remote job in parameter.
    """
    if job.id is None:
        raise ValueError("Can't retrieve remote result for a job whose id is None.")
    return remote_result_from_id(job.id, job.device)


def get_all_job_ids() -> dict[type[AvailableDevice], list[str]]:
    """Retrieve from the remote providers all the job-ids associated with this
    account.

    Returns:
        A dictionary of job-ids indexed by the correspond AvailableDevice
        (ATOSDevice, AWSDevice, IBMDevice, ...).
    """
    job_ids: dict[type[AvailableDevice], list[str]] = {
        AWSDevice: get_all_task_ids(),
        ATOSDevice: qlm_ids(),
        IBMDevice: ibm_ids(),
    }

    return job_ids



# %%
import re

# import cma
import numpy as np
import scipy.special as ss
from matplotlib import pyplot as plt
from math import floor
from scipy.optimize import minimize

from mpqp.all import *


# %%
def Cheb(_x, _n):
    return np.cos(_n * np.arccos(_x))


def Chebyshev(_x, _n, _nOrder):
    der = 0
    if _nOrder == 0:
        return Cheb(_x, _n)
    else:
        for k in range((_n - _nOrder) % 2, _n - _nOrder + 1, 2):
            binomialFactor = ss.binom(
                (_n + _nOrder - k) / 2 - 1, (_n - _nOrder - k) / 2
            )
            factor = ss.factorial((_n + _nOrder + k) / 2 - 1)
            denominator = ss.factorial((_n - _nOrder + k) / 2)
            if k == 0:
                der += binomialFactor * factor / (2 * denominator) * Cheb(_x, k)
            else:
                der += binomialFactor * factor / denominator * Cheb(_x, k)
        return 2**_nOrder * _n * der


# %%
def diagonal_observable_cheb(_x, _nQubits, _nOrder):
    diagonal_elements_plus = np.array(
        [Chebyshev(_x, index, _nOrder) for index in range(2 ** (_nQubits - 1))]
    )
    diagonal_elements_minus = -diagonal_elements_plus

    diagonal_elements = np.concatenate(
        (diagonal_elements_plus, diagonal_elements_minus), axis=None
    )
    # print(diagonal_elements)
    return Operator(np.diag(v=diagonal_elements))


# %%
def vqc(list_parameters, _depth, _nQubits):
    """
    Generates a VQC of depth d, for N qubits, with list_parameters as angles.

    :param list_parameters: LIST or np.ARRAY of OPTIMIZABLE parameters. It should have length 3Nd elements in [0,2pi]
    :param N: INTEGER, number of qubits
    :param d: INTEGER, depth of the vqc
    :return: a circuit corresponding to (vqa)|0>
    """
    reg = QuantumRegister(_nQubits)
    _qc = QuantumCircuit(reg)

    for d in range(_depth):
        for index, r in enumerate(reg):
            _qc.ry((list_parameters[_nQubits * d + index]), r)

        [_qc.cx(reg[k], reg[k + 1]) for k in range(_nQubits - 1)]

    return _qc


# %%
def extractOrders(_expressions):  # works up to order 9
    functions = ["f", "g", "h", "k"][: len(_expressions)]

    orders = []
    for expression in _expressions:
        orderFunctions = []
        for function in functions:
            listDerivatives = re.findall(r"d+\w" + function, expression)
            listDerivatives.sort(reverse=True)
            listDerivatives = [
                derivative.replace(function, "") for derivative in listDerivatives
            ]
            listDerivatives = [
                derivative.replace("d", "") for derivative in listDerivatives
            ]
            if len(listDerivatives) == 0:
                orderFunctions.append(0)
            else:
                orderFunctions.append(int(listDerivatives[0]))
        orders.append(orderFunctions)
    orders = np.array(orders)

    maxOrderFunctions = [max(orders[:, i]) for i in range(len(_expressions))]
    return maxOrderFunctions


# %%
nb_qubits = 2
depth = 2
nb_params = nb_qubits * depth

n_samples, x_min, x_max = 20, 0, 0.95
x_samples = np.linspace(x_min, x_max, n_samples)

epsilon_0 = 0.1
sigma_0 = 5
n = 4
b = 10
K = 100
# expressions = ['d1f - d0g/(3*K) - (2/np.sqrt(3))*epsilon_0*(d0g/(np.sqrt(3)*sigma_0))**n', 'd1g + b']
l = 8
k = 0.1
# expressions = ["d1f + l*d0f*(k + np.tan(l*x))"]
expressions = ["d1f - 1"]


functions = ["f", "g", "h", "k"][: len(expressions)]
orders = extractOrders(expressions)
boundaries = {"d0f": [0.0, 1.0]}
# boundaries = {"d0f": [0.0, 1.0], "d0g": [0.95, 2.]}

_nb_variables = len(functions) * nb_params + len(functions)

shots = None


# %%
def expectationValues(estimator, _attemptParameters, _shots=None):
    scaling_factor = _attemptParameters[: len(functions)]
    listExpectationValues = []
    for indexFunction in range(len(functions)):
        gamma = scaling_factor[indexFunction]
        angles = _attemptParameters[
            len(functions)
            + indexFunction * nb_params : len(functions)
            + (indexFunction + 1) * nb_params
        ]
        circ = vqc(angles, depth, nb_qubits)

        for order in range(orders[indexFunction] + 1):
            for sample in x_samples:
                observable = gamma * diagonal_observable_cheb(sample, nb_qubits, order)
                psi = circ
                job = estimator.run([psi], [observable], shots=_shots)
                listExpectationValues.append(job.result().values[0])

    return listExpectationValues


# %%
def cost_function(_attemptParameters, estimator):
    _expectationValues = expectationValues(estimator, _attemptParameters, shots)
    i = 0
    for indexFunction in range(len(functions)):
        for order in range(orders[indexFunction] + 1):
            locals()["_d" + str(order) + functions[indexFunction]] = np.array(
                _expectationValues[i * n_samples : (i + 1) * n_samples]
            )
            i += 1

    for derivative, coordinate in boundaries.items():
        indexCoordinate = floor(coordinate[0] * n_samples / (x_max - x_min))
        if indexCoordinate == n_samples:
            indexCoordinate -= 1
        shift = coordinate[1] - locals()["_" + derivative][indexCoordinate]
        for indexFunction in range(len(functions)):
            for order in range(orders[indexFunction] + 1):
                if derivative.find("d" + str(order) + functions[indexFunction]) != -1:
                    locals()["_d" + str(order) + functions[indexFunction]] += shift

    loss = 0
    for index in range(n_samples):
        x = x_samples[index]
        for expression in expressions:
            for indexFunction in range(len(functions)):
                for order in range(orders[indexFunction] + 1):
                    if (
                        expression.find("d" + str(order) + functions[indexFunction])
                        != -1
                    ):
                        locals()[
                            "d" + str(order) + functions[indexFunction]
                        ] = locals()["_d" + str(order) + functions[indexFunction]][
                            index
                        ]
        for expression in expressions:
            loss += abs(eval(expression)) ** 2

    return loss  # /n_samples #+ loss_BC


# %%
init_theta_list = np.random.uniform(low=0, high=2 * 3.14, size=(_nb_variables,))

# %%
# _options = cma.CMAOptions()
# _options.set('tolfun', 1e-3)
# opt = cma.fmin(cost_function, x0=init_theta_list, sigma0=1.5, options = _options)

method = "BFGS"
options = {"disp": True, "maxiter": 10}
# def callback(param_list): #LOCAL
#     current_loss = cost_function(param_list) #LOCAL
#     # loss_iterations.append(current_loss)    #LOCAL
#     print("current loss: ",current_loss,"\n") #LOCAL

# estimator=Estimator() #LOCAL

service = QiskitRuntimeService()  # IBMQ REMOTE
session = Session(service, backend="ibmq_qasm_simulator")  # IBMQ REMOTE
estimator = Estimator(session=session, options=None)  # IBMQ REMOTE

opt = minimize(
    cost_function, init_theta_list, args=(estimator), method=method, options=options
)
# %%
parameters = opt.x

# %%
scaling_factor = parameters[: len(functions)]
for index in range(len(functions)):
    angles = parameters[
        len(functions) + index * nb_params : len(functions) + (index + 1) * nb_params
    ]
    circ = vqc(angles, depth, nb_qubits)

    solution_attempt = []

    for sample in x_samples:
        observable = scaling_factor[index] * diagonal_observable_cheb(
            sample, nb_qubits, 0
        )
        psi = circ
        job = estimator.run([psi], [observable])
        locals()[functions[index]] = job.result().values[0]
        solution_attempt.append(locals()[functions[index]])

    solution_attempt = np.array(solution_attempt)
    indexCoordinate = int(
        floor(boundaries["d0" + functions[index]][0] * n_samples / (x_max - x_min))
    )
    shift = boundaries["d0" + functions[index]][1] - solution_attempt[indexCoordinate]

    solution_attempt += shift
    locals()[functions[index]] = solution_attempt


# %%

# def target_function_f(_x):
#     return np.exp(-0.1*8*_x) * np.cos(8*_x)


# def target_function_f(_x):
#     return _x


# def target_function_f(_x):
#     # return np.exp(_x)
#     return np.exp(-k * _x * l) * np.cos(l * _x)
def target_function_f(_x):
    return _x + 1


# def target_function_g(_x):
#     return (1/5)*( - 5*_x - 10*(_x**2) - 10*(_x**3) -5*(_x**4) - (_x**5) )

# solutions MD
# def target_function_f(_x):
#      return x_samples*(0.337217 - 0.563122*x_samples + 0.496778*(x_samples**2) - 0.225808*(x_samples**3) + 0.041056*(x_samples**4))
# def target_function_g(_x):
#      return 11 - 10*x_samples

for function in functions:
    plt.plot(
        x_samples,
        locals()["target_function_" + function](x_samples),
        linewidth=4,
        label="Target",
        color="Blue",
    )

    plt.ylabel("y")
    plt.plot(x_samples, locals()[function], label="Attempt OA", color="orange")
    plt.legend()
    plt.savefig(
        "TAN_CMA_" + str(nb_qubits) + "q" + str(depth) + "d" + str(shots) + "tol-2.png",
        dpi=300,
    )

# %%
# session.close() #IBMQ REMOTE



# pyright: reportUnusedImport=false
from mpqp.core.circuit import QCircuit
from mpqp.core.instruction.barrier import Barrier
from mpqp.core.languages import Language
from mpqp.core.instruction import Instruction



from __future__ import annotations

from typing import Any, Collection, TypeVar, Union
from typing import Callable, Optional
from scipy.optimize import minimize as scipy_minimize, OptimizeResult
from sympy import Expr
import numpy as np
import numpy.typing as npt
from typeguard import typechecked

from mpqp.core.circuit import QCircuit
from mpqp.execution.devices import AvailableDevice
from mpqp.execution.runner import _run_single  # pyright: ignore[reportPrivateUsage]
from mpqp.execution.vqa.optimizer import Optimizer

T1 = TypeVar("T1")
T2 = TypeVar("T2")
OptimizerInput = Union[list[float], npt.NDArray[np.float32]]
OptimizableFunc = Callable[[OptimizerInput], float]
OptimizerCallable = Callable[
    [OptimizableFunc, Optional[OptimizerInput]], tuple[float, OptimizerInput]
]

# 3M-TODO: all those functions with almost or exactly the same signature look like
#  a code smell to me.

# 3M-TODO: test the minimizer options


def _maps(l1: Collection[T1], l2: Collection[T2]) -> dict[T1, T2]:
    """Does like zip, but with a dictionary instead of a list of tuples"""
    assert len(l1) == len(l2)
    return {e1: e2 for e1, e2 in zip(l1, l2)}


@typechecked
def minimize(
    optimizable: QCircuit | OptimizableFunc,
    method: Optimizer | OptimizerCallable,
    device: Optional[AvailableDevice] = None,
    init_params: Optional[OptimizerInput] = None,
    nb_params: Optional[int] = None,
    optimizer_options: Optional[dict[str, Any]] = None,
) -> tuple[float, OptimizerInput]:
    """This function runs an optimization on the parameters of the circuit, in order to
    minimize the measured expectation value of observables associated with the given circuit.
    Note that this means that the latter should contain an ``ExpectationMeasure``.

    Examples:
        >>> alpha, beta = symbols("α β")
        >>> circuit = QCircuit([
        ...     H(0),
        ...     Rx(alpha, 1),
        ...     CNOT(1,0),
        ...     Rz(beta, 0),
        ...     ExpectationMeasure(
        ...         [0,1],
        ...         observable=Observable(np.diag([1,2,-3,4])),
        ...         shots=0,
        ...     ),
        ... ])
        >>> minimize(
        ...     circuit,
        ...     Optimizer.BFGS,
        ...     ATOSDevice.MYQLM_PYLINALG,
        ...     optimizer_options={"maxiter":50},
        ... )
        (-0.9999999999999996, array([0., 0.]))


        >>> def cost_func(params):
        ...     run_res = run(
        ...         circuit,
        ...         ATOSDevice.MYQLM_PYLINALG,
        ...         {alpha: params[0], beta: params[1]}
        ...     )
        ...     return 1 - run_res.expectation_value ** 2
        >>> minimize(
        ...     cost_func,
        ...     Optimizer.BFGS,
        ...     nb_params=2,
        ...     optimizer_options={"maxiter":50},
        ... )
        (8.881784197001252e-16, array([0., 0.]))

    Args:
        optimizable: Either the circuit, containing symbols and an expectation
            measure, or the evaluation function.
        method: The method used to optimize most of those methods come from
            ``scipy``. If the choices offered in this package are not
            covering your needs, you can define your own optimizer. This should be
            a function taking as input a function representing the circuit, with
            as many inputs as the circuit has parameters, and any optional
            initialization parameters, and returning the optimal value reached
            and the parameters used to reach this value.
        device: The device on which the circuit should be run.
        init_params: The optional initialization parameters (the value
            attributed to the symbols in the first loop of the optimizer).
        nb_params: Number of variables to input in ``optimizable``. It is only
            useful if ``optimizable`` is a Callable and if ``init_params`` was
            not given. If not this argument is not taken into account.
        optimizer_options: Options used to configure the VQA optimizer (maximum
            iterations, convergence threshold, etc...). These options are passed
            as is to the minimizer.

    Returns:
        The optimal value reached and the parameters corresponding to this value.
    """
    if isinstance(optimizable, QCircuit):
        if device is None:
            raise ValueError("A device is needed to optimize a circuit")
        optimizer = _minimize_remote if device.is_remote() else _minimize_local
        return optimizer(optimizable, method, device, init_params, nb_params)
    else:
        # 3M-TODO: find a way to know if the job is remote or local from the function
        return _minimize_local(
            optimizable, method, device, init_params, nb_params, optimizer_options
        )


@typechecked
def _minimize_remote(
    optimizable: QCircuit | OptimizableFunc,
    method: Optimizer | OptimizerCallable,
    device: Optional[AvailableDevice] = None,
    init_params: Optional[OptimizerInput] = None,
    nb_params: Optional[int] = None,
    optimizer_options: Optional[dict[str, Any]] = None,
) -> tuple[float, OptimizerInput]:
    """This function runs an optimization on the parameters of the circuit, to
    minimize the expectation value of the measure of the circuit by it's
    observables. Note that this means that the circuit should contain an
    expectation measure.

    Args:
        optimizable: Either the circuit, containing symbols and an expectation
            measure, or the evaluation function.
        method: The method used to optimize most of those methods come from
            either scipy or cma. If the choice offered in this package are not
            covering your needs, you can define your own optimizer. It should be
            a function taking as input a function representing the circuit, with
            as many inputs as the circuit has parameters, as well as optional
            initialization parameters, and returning the optimal value reached
            and the parameters used to reach this value.
        device: The device on which the circuit should be run.
        init_params: The optional initialization parameters (the value
            attributed to the symbols in the first loop of the optimizer).
        nb_params: number of variables to input in ``optimizable``. It is only
            useful if ``optimizable`` is a Callable and if ``init_params`` was
            not given. If not this argument is not taken into account.
        optimizer_options: Options used to configure the VQA optimizer (maximum
            iterations, convergence threshold, etc...). These options are passed
            as is to the minimizer.

    Returns:
        The optimal value reached and the parameters used to reach this value.

    3M-TODO to implement on QLM first
    """
    raise NotImplementedError()


@typechecked
def _minimize_local(
    optimizable: QCircuit | OptimizableFunc,
    method: Optimizer | OptimizerCallable,
    device: Optional[AvailableDevice] = None,
    init_params: Optional[OptimizerInput] = None,
    nb_params: Optional[int] = None,
    optimizer_options: Optional[dict[str, Any]] = None,
) -> tuple[float, OptimizerInput]:
    """This function runs an optimization on the parameters of the circuit, to
    minimize the expectation value of the measure of the circuit by it's
    observables. Note that this means that the circuit should contain an
    expectation measure.

    Args:
        optimizable: Either the circuit, containing symbols and an expectation
            measure, or the evaluation function.
        method: The method used to optimize most of those methods come from
            either scipy or cma. If the choice offered in this package are not
            covering your needs, you can define your own optimizer. It should be
            a function taking as input a function representing the circuit, with
            as many inputs as the circuit has parameters, as well as optional
            initialization parameters, and returning the optimal value reached
            and the parameters used to reach this value.
        device: The device on which the circuit should be run.
        init_params: The optional initialization parameters (the value
            attributed to the symbols in the first loop of the optimizer).
        nb_params: number of variables to input in ``optimizable``. It is only
            useful if ``optimizable`` is a Callable and if ``init_params`` was
            not given. If not this argument is not taken into account.
        optimizer_options: Options used to configure the VQA optimizer (maximum
            iterations, convergence threshold, etc...). These options are passed
            as is to the minimizer.

    Returns:
        the optimal value reached and the parameters used to reach this value.
    """
    if isinstance(optimizable, QCircuit):
        if device is None:
            raise ValueError("A device is needed to optimize a circuit")
        return _minimize_local_circ(
            optimizable, device, method, init_params, optimizer_options
        )
    else:
        return _minimize_local_func(
            optimizable, method, init_params, nb_params, optimizer_options
        )


@typechecked
def _minimize_local_circ(
    circ: QCircuit,
    device: AvailableDevice,
    method: Optimizer | OptimizerCallable,
    init_params: Optional[OptimizerInput] = None,
    optimizer_options: Optional[dict[str, Any]] = None,
) -> tuple[float, OptimizerInput]:
    """This function runs an optimization on the parameters of the circuit, to
    minimize the expectation value of the measure of the circuit by it's
    observables. Note that this means that the circuit should contain an
    expectation measure!

    Args:
        circ: Either the circuit, containing symbols and an expectation measure.
        method: The method used to optimize most of those methods come from
            either scipy or cma. If the choice offered in this package are not
            covering your needs, you can define your own optimizer. It should be
            a function taking as input a function representing the circuit, with
            as many inputs as the circuit has parameters, as well as optional
            initialization parameters, and returning the optimal value reached
            and the parameters used to reach this value.
        device: The device on which the circuit should be run.
        init_params: The optional initialization parameters (the value
            attributed to the symbols in the first loop of the optimizer).
        optimizer_options: Options used to configure the VQA optimizer (maximum
            iterations, convergence threshold, etc...). These options are passed
            as is to the minimizer.

    Returns:
        The optimal value reached and the parameters used to reach this value.
    """
    # The sympy `free_symbols` method returns in fact sets of Basic, which
    # are theoretically different from Expr, but in our case the difference
    # is not relevant.
    variables: set[Expr] = circ.variables()  # type: ignore

    def eval_circ(params: OptimizerInput):
        # pyright is bad with abstract numeric types:
        # "float" is incompatible with "Complex"
        return _run_single(
            circ,
            device,
            _maps(variables, params),  # pyright: ignore[reportArgumentType]
        ).expectation_value

    return _minimize_local_func(
        eval_circ, method, init_params, len(variables), optimizer_options
    )


@typechecked
def _minimize_local_func(
    eval_func: OptimizableFunc,
    method: Optimizer | OptimizerCallable,
    init_params: Optional[OptimizerInput] = None,
    nb_params: Optional[int] = None,
    optimizer_options: Optional[dict[str, Any]] = None,
) -> tuple[float, OptimizerInput]:
    """This function runs an optimization on the parameters of the circuit, to
    minimize the expectation value of the measure of the circuit by it's
    observables. Note that this means that the circuit should contain an
    expectation measure!

    Args:
        eval_func: Evaluation function.
        method: The method used to optimize most of those methods come from
            either scipy or cma. If the choice offered in this package are not
            covering your needs, you can define your own optimizer. It should be
            a function taking as input a function representing the circuit, with
            as many inputs as the circuit has parameters, as well as optional
            initialization parameters, and returning the optimal value reached
            and the parameters used to reach this value.
        init_params: The optional initialization parameters (the value
            attributed to the symbols in the first loop of the optimizer).
        nb_params: number of variables to input in ``optimizable``. It is only
            useful if ``init_params`` was not given. If not this argument is not
            taken into account.
        optimizer_options: Options used to configure the VQA optimizer (maximum
            iterations, convergence threshold, etc...). These options are passed
            as is to the minimizer.

    Returns:
        The optimal value reached and the parameters used to reach this value.
    """
    if init_params is None:
        if nb_params is None:
            raise ValueError(
                "Please provide either a set of initialization parameters or "
                "the number of parameters expected by the function."
            )
        else:
            init_params = [0.0] * nb_params

    if isinstance(method, Optimizer):
        assert method is not None
        res: OptimizeResult = scipy_minimize(
            eval_func,
            x0=np.array(init_params),
            method=method.name.lower(),
            options=optimizer_options,
        )
        return res.fun, res.x
    else:
        return method(eval_func, init_params)
